@@ -38,9 +38,6 @@
 #include "gettext.h"
 #define _(str) gettext (str)
 
-/* Buffer size for holding a file name.  FIXME: No fix limit in GNU... */
-#define NAME_BUFFER_SIZE 1024
-
 /* Buffer size for shell process input.  */
 #define SHELL_BUFFER_SIZE 8196
 
@@ -76,6 +73,7 @@ static char *current_directory;
 #define starting_with(data, pattern)					    \
   (memcmp (data, pattern, sizeof (pattern) - 1) == 0)
 
+static FILE* load_stdin (char **p_name_buf);
 
 /*-------------------------------------------------------------------------.
 | For a DATA string and a PATTERN containing one or more embedded	   |
@@ -359,10 +357,8 @@ main (argc, argv)
      int argc;
      char *const *argv;
 {
-  size_t size_read;
   FILE *file;
-  char name_buffer[NAME_BUFFER_SIZE];
-  char copy_buffer[NAME_BUFFER_SIZE];
+  char* name_buffer;
   int optchar;
 
   program_name = argv[0];
@@ -420,7 +416,7 @@ main (argc, argv)
 This is free software; see the source for copying conditions.  There is NO\n\
 warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 "),
-	      "1994, 1995, 1996");
+	      "1994, 1995, 1996, 2005");
       exit (EXIT_SUCCESS);
     }
 
@@ -428,44 +424,123 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
     usage (EXIT_SUCCESS);
 
   if (optind < argc)
-    for (; optind < argc; optind++)
-      {
-	if (argv[optind][0] == '/')
-	  stpcpy (name_buffer, argv[optind]);
-	else
-	  {
-	    char *cp = stpcpy (name_buffer, current_directory);
-	    *cp++ = '/';
-	    stpcpy (cp, argv[optind]);
-	  }
-	if (file = fopen (name_buffer, "r"), !file)
-	  error (EXIT_FAILURE, errno, name_buffer);
-	unarchive_shar_file (name_buffer, file);
-	fclose (file);
-      }
+    {
+      size_t buflen = 0;
+      argv += optind;
+
+      for (; optind < argc; optind++)
+        {
+          char* arg = *(argv++);
+          size_t sz = strlen (current_directory) + strlen (argv[optind]) + 2;
+          if (sz > buflen)
+            {
+              buflen = sz + 32;
+              name_buffer = (name_buffer == NULL)
+                ? malloc (buflen)
+                : realloc (name_buffer, buflen);
+              if (name_buffer == NULL)
+                error (EXIT_FAILURE, ENOMEM, _("allocate file name buffer"));
+            }
+
+          if (argv[optind][0] == '/')
+            strcpy (name_buffer, argv[optind]);
+          else
+            {
+              char *cp = stpcpy (name_buffer, current_directory);
+              *cp++ = '/';
+              strcpy (cp, argv[optind]);
+            }
+          if (file = fopen (name_buffer, "r"), !file)
+            error (EXIT_FAILURE, errno, name_buffer);
+          unarchive_shar_file (name_buffer, file);
+          fclose (file);
+        }
+
+    }
   else
     {
-      sprintf (name_buffer, "/tmp/unsh.%05d", (int) getpid ());
-      unlink (name_buffer);
-
-      if (file = fopen (name_buffer, "w+"), !file)
-	error (EXIT_FAILURE, errno, name_buffer);
-#ifndef __MSDOS__
-      unlink (name_buffer);	/* will be deleted on fclose */
-#endif
-
-      while (size_read = fread (copy_buffer, 1, sizeof (copy_buffer), stdin),
-	     size_read != 0)
-	fwrite (copy_buffer, size_read, 1, file);
-      rewind (file);
+      file = load_stdin (&name_buffer);
 
       unarchive_shar_file (_("standard input"), file);
 
       fclose (file);
-#ifdef __MSDOS__
-      unlink (name_buffer);
-#endif
     }
 
   exit (EXIT_SUCCESS);
 }
+
+
+static FILE*
+load_stdin (char **p_name_buf)
+{
+  static const char z_tmpfile[] = "unsh.XXXXXX";
+  char *pz_fname;
+  FILE *fp;
+
+  /*
+   * FIXME: actually configure this stuff:
+   */
+#if defined(_SC_PAGESIZE)
+  long pg_sz = sysconf (_SC_PAGESIZE);
+#elif defined(_SC_PAGE_SIZE)
+  long pg_sz = sysconf (_SC_PAGE_SIZE);
+#elif defined(HAVE_GETPAGESIZE)
+  long pg_sz = getpagesize();
+#else
+# define pg_sz 8192
+#endif
+
+  {
+    size_t name_size;
+    char *pz_tmp = getenv ("TMPDIR");
+
+    if (pz_tmp == NULL)
+      pz_tmp = "/tmp";
+
+    name_size = strlen (pz_tmp) + sizeof (z_tmpfile) + 1;
+    *p_name_buf = pz_fname = malloc (name_size);
+
+    if (pz_fname == NULL)
+      error (EXIT_FAILURE, ENOMEM, _("allocate file name buffer"));
+
+    sprintf (pz_fname, "%s/%s", pz_tmp, z_tmpfile);
+  }
+
+  {
+    int fd = mkstemp (pz_fname);
+    if (fd < 0)
+      error (EXIT_FAILURE, errno, pz_fname);
+
+    fp = fdopen (fd, "w+");
+  }
+
+  if (fp == NULL)
+    error (EXIT_FAILURE, errno, pz_fname);
+
+  {
+    char *buf = malloc (pg_sz);
+    size_t size_read;
+
+    if (buf == NULL)
+      error (EXIT_FAILURE, ENOMEM, _("allocate file buffer"));
+
+    while (size_read = fread (buf, 1, pg_sz, stdin),
+           size_read != 0)
+      fwrite (buf, size_read, 1, fp);
+
+    free (buf);
+  }
+
+  rewind (fp);
+
+  return fp;
+}
+
+/*
+ * Local Variables:
+ * mode: C
+ * c-file-style: "gnu"
+ * tab-width: 8
+ * indent-tabs-mode: nil
+ * End:
+ * end of agen5/autogen.c */
