@@ -1,5 +1,5 @@
 /* uudecode utility.
-   Copyright (C) 1994, 1995 Free Software Foundation, Inc.
+   Copyright (C) 1994, 1995, 1996, 2002 Free Software Foundation, Inc.
 
    This product is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -49,7 +49,23 @@
 
 /* Reworked to GNU style by Ian Lance Taylor, ian@airs.com, August 93.  */
 
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+#include "liballoca.h"
+
 #include "system.h"
+#include "basename.h"
+#include "error.h"
+#include "exit.h"
+
+#if HAVE_LOCALE_H
+# include <locale.h>
+#else
+# define setlocale(Category, Locale)
+#endif
+#include "gettext.h"
+#define _(str) gettext (str)
 
 /*=====================================================================\
 | uudecode [FILE ...]						       |
@@ -70,10 +86,14 @@ static struct option longopts[] =
   { NULL, 0, NULL, 0 }
 };
 
-static int read_stduu __P ((const char *inname));
-static int read_base64 __P ((const char *inname));
+static int read_stduu  __P ((const char *inname, const char *outname));
+static int read_base64 __P ((const char *inname, const char *outname));
 static int decode __P ((const char *, const char *));
-static void usage __P ((int));
+static void usage __P ((int))
+#if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ >= 5) || __GNUC__ > 2)
+     __attribute__ ((noreturn))
+#endif
+;
 
 /* The name this program was run with. */
 const char *program_name;
@@ -81,10 +101,20 @@ const char *program_name;
 /* Single character decode.  */
 #define	DEC(Char) (((Char) - ' ') & 077)
 
+#if !defined S_ISLNK && defined S_IFLNK
+# define S_ISLNK(m) (((m) & S_IFMT) == S_IFLNK)
+#endif
+
+#define TRY_PUTCHAR(c) do { \
+                         if (putchar (c) == EOF) { \
+                           error (0, 0, _("%s: Write error"), outname); \
+                           return 1; \
+                         } \
+                       } while (0)
 
 static int
-read_stduu (inname)
-     const char *inname;
+read_stduu (inname, outname)
+     const char *inname, *outname;
 {
   char buf[2 * BUFSIZ];
 
@@ -93,7 +123,7 @@ read_stduu (inname)
       int n;
       char *p;
 
-      if (fgets (buf, sizeof(buf), stdin) == NULL)
+      if (fgets ((char *) buf, sizeof(buf), stdin) == NULL)
 	{
 	  error (0, 0, _("%s: Short file"), inname);
 	  return 1;
@@ -106,32 +136,17 @@ read_stduu (inname)
       n = DEC (*p);
       if (n <= 0)
 	break;
-      for (++p; n > 0; p += 4, n -= 3)
+      for (++p; n >= 3; p += 4, n -= 3)
 	{
-	  char ch;
-
-	  if (n >= 3)
-	    {
-	      ch = DEC (p[0]) << 2 | DEC (p[1]) >> 4;
-	      putchar (ch);
-	      ch = DEC (p[1]) << 4 | DEC (p[2]) >> 2;
-	      putchar (ch);
-	      ch = DEC (p[2]) << 6 | DEC (p[3]);
-	      putchar (ch);
-	    }
-	  else
-	    {
-	      if (n >= 1)
-		{
-		  ch = DEC (p[0]) << 2 | DEC (p[1]) >> 4;
-		  putchar (ch);
-		}
-	      if (n >= 2)
-		{
-		  ch = DEC (p[1]) << 4 | DEC (p[2]) >> 2;
-		  putchar (ch);
-		}
-	    }
+	  TRY_PUTCHAR (DEC (p[0]) << 2 | DEC (p[1]) >> 4);
+	  TRY_PUTCHAR (DEC (p[1]) << 4 | DEC (p[2]) >> 2);
+	  TRY_PUTCHAR (DEC (p[2]) << 6 | DEC (p[3]));
+        }
+      if (n > 0)
+	{
+	  TRY_PUTCHAR (DEC (p[0]) << 2 | DEC (p[1]) >> 4);
+	  if (n >= 2)
+	    TRY_PUTCHAR (DEC (p[1]) << 4 | DEC (p[2]) >> 2);
 	}
     }
 
@@ -146,8 +161,8 @@ read_stduu (inname)
 }
 
 static int
-read_base64 (inname)
-     const char *inname;
+read_base64 (inname, outname)
+     const char *inname, *outname;
 {
   static const char b64_tab[256] =
   {
@@ -184,7 +199,7 @@ read_base64 (inname)
     '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*360-367*/
     '\177', '\177', '\177', '\177', '\177', '\177', '\177', '\177', /*370-377*/
   };
-  unsigned char buf[2 * BUFSIZ];
+  char buf[2 * BUFSIZ];
 
   while (1)
     {
@@ -196,7 +211,7 @@ read_base64 (inname)
 	  error (0, 0, _("%s: Short file"), inname);
 	  return 1;
 	}
-      p = buf;
+      p = (unsigned char *) buf;
 
       if (memcmp (buf, "====", 4) == 0)
 	break;
@@ -239,7 +254,7 @@ read_base64 (inname)
 	      }
 	  if (*p == '=')
 	    {
-	      putchar (c1 << 2 | c2 >> 4);
+	      TRY_PUTCHAR (c1 << 2 | c2 >> 4);
 	      last_data = 1;
 	      break;
 	    }
@@ -251,15 +266,15 @@ read_base64 (inname)
 		error (0, 0, _("%s: illegal line"), inname);
 		return 1;
 	      }
-	  putchar (c1 << 2 | c2 >> 4);
-	  putchar (c2 << 4 | c3 >> 2);
+	  TRY_PUTCHAR (c1 << 2 | c2 >> 4);
+	  TRY_PUTCHAR (c2 << 4 | c3 >> 2);
 	  if (*p == '=')
 	    {
 	      last_data = 1;
 	      break;
 	    }
 	  else
-	    putchar (c3 << 6 | b64_tab[*p++]);
+	    TRY_PUTCHAR (c3 << 6 | b64_tab[*p++]);
 	}
     }
 
@@ -279,6 +294,8 @@ decode (inname, forced_outname)
   char buf[2 * BUFSIZ];
   char *outname;
   int do_base64 = 0;
+  struct stat attr;
+  int rval;
 
   /* Search for header line.  */
 
@@ -292,12 +309,12 @@ decode (inname, forced_outname)
 
       if (strncmp (buf, "begin", 5) == 0)
 	{
-	  if (sscanf (buf, "begin-base64 %o %s", &mode, buf) == 2)
+	  if (sscanf (buf, "begin-base64 %o %[^\n]", &mode, buf) == 2)
 	    {
 	      do_base64 = 1;
 	      break;
 	    }
-	  else if (sscanf (buf, "begin %o %s", &mode, buf) == 2)
+	  else if (sscanf (buf, "begin %o %[^\n]", &mode, buf) == 2)
 	    break;
 	}
     }
@@ -337,6 +354,23 @@ decode (inname, forced_outname)
 	}
     }
 
+  /* Check out file if it exists */
+  if (strcmp (outname, "/dev/stdout") != 0 && strcmp (outname, "-") != 0
+      && !access(outname, F_OK)) {
+    if (lstat(outname, &attr) == -1) {
+      error (0, errno, _("cannot access %s"), outname);
+      return 1;
+    }
+    if (S_ISFIFO(attr.st_mode)){
+      error (0, errno, _("denied writing FIFO (%s)"), outname);
+      return 1;
+    }
+    if (S_ISLNK(attr.st_mode)) {
+      error (0, errno, _("not following symlink (%s)"), outname);
+      return 1;
+    }
+  }
+
   /* Create output file and set mode.  */
 
   if (strcmp (outname, "/dev/stdout") != 0 && strcmp (outname, "-") != 0
@@ -357,9 +391,17 @@ decode (inname, forced_outname)
 
   /* For each input line:  */
   if (do_base64)
-    return read_base64 (inname);
+    rval = read_base64 (inname, outname);
   else
-    return read_stduu (inname);
+    rval = read_stduu (inname, outname);
+
+  if (rval == 0 && (ferror(stdout) || fflush(stdout) != 0))
+    {
+      error (0, 0, _("%s: Write error"), outname);
+      return 1;
+    }
+
+  return rval;
 }
 
 static void
@@ -372,11 +414,13 @@ usage (status)
   else
     {
       printf (_("Usage: %s [FILE]...\n"), program_name);
-      printf (_("\
+      fputs (_("\
 Mandatory arguments to long options are mandatory to short options too.\n\
-  -h, --help               display this help and exit\n\
-  -v, --version            output version information and exit\n\
-  -o, --output-file=FILE   direct output to FILE\n"));
+  -o, --output-file=FILE   direct output to FILE\n\
+      --help               display this help and exit\n\
+      --version            output version information and exit\n"), stdout);
+      /* TRANSLATORS: add the contact address for your translation team! */
+      printf (_("Report bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
   exit (status);
 }
@@ -398,7 +442,7 @@ main (argc, argv)
   bindtextdomain (PACKAGE, LOCALEDIR);
   textdomain (PACKAGE);
 
-  while (opt = getopt_long (argc, argv, "ho:v", longopts, (int *) NULL),
+  while (opt = getopt_long (argc, argv, "o:", longopts, (int *) NULL),
 	 opt != EOF)
     {
       switch (opt)
@@ -411,7 +455,14 @@ main (argc, argv)
 	  break;
 
 	case 'v':
-	  printf ("%s - GNU %s %s\n", program_name, PACKAGE, VERSION);
+	  printf ("%s (GNU %s) %s\n", basename (program_name),
+		  PACKAGE, VERSION);
+	  /* xgettext: no-wrap */
+	  printf (_("Copyright (C) %s Free Software Foundation, Inc.\n\
+This is free software; see the source for copying conditions.  There is NO\n\
+warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
+"),
+		  "1994, 1995, 1996, 2002");
 	  exit (EXIT_SUCCESS);
 
 	case 0:
